# Author: Swati Swati (swati17293@gmail.com, swati.swati@unibw.de)
"""
Fairness Visualization Report Generator
This module generates a detailed fairness report using the Fairlearn library, providing both group-wise and scalar fairness metrics across sensitive attributes such as sex or race.
"""

from mammoth_commons.datasets import CSV
from mammoth_commons.exports import HTML
from mammoth_commons.models import Predictor
from mammoth_commons.integration import metric

from typing import List


@metric(
    namespace="mammotheu",
    version="v0001",
    python="3.13",
    packages=("fairlearn", "plotly", "pandas", "onnxruntime", "mmm-fair"),
)
def viz_fairness_report(
    dataset: CSV,
    model: Predictor,
    sensitive: List[str],
) -> HTML:
    """
    <p>This module generates a detailed fairness report using <a href="https://fairlearn.org/" target="_blank">Fairlearn</a>,
    providing both group-wise and scalar fairness metrics across sensitive attributes such as sex or race.</p>
    """
    # Import the existing function from mmm-fair
    from mmm_fair.fairlearn_report import generate_reports_from_fairlearn

    # Extract predictions and ground truth
    if hasattr(model, "mmm"):
        model = model.mmm
    y_true = list(dataset.labels.values())[-1]
    y_pred = model.predict(dataset, sensitive)

    # Convert sensitive attributes into the expected numpy array format
    import numpy as np

    sa_df = dataset.data[sensitive].copy()
    sa_matrix = sa_df.to_numpy()

    # Force the first group in each sensitive column to be privileged (0), rest as 1
    for col_idx, attr in enumerate(sensitive):
        col = sa_matrix[:, col_idx]
        first_group = col[0]  # Treat first seen value as privileged
        sa_matrix[:, col_idx] = np.where(col == first_group, 0, 1)

    sa_matrix = sa_matrix.astype(int)

    # Call the original report function from mmm-fair (already supports console/table/html)
    html_string = generate_reports_from_fairlearn(
        report_type="table",  # one could also use "table" if preferred
        sensitives=sensitive,
        mmm_classifier=model,
        saIndex_test=sa_matrix,
        y_pred=y_pred,
        y_test=y_true,
        launch_browser=False,  # suppresses opening in a new browser
    )

    return HTML(html_string)
